/*--------------------------------------------------------------------
  adc_temp_led: based on assembler "adc3" project, but uses 3 digit
                7 segment LED display instead of LCD.

				This version uses the Microchip TC1046 temperature sensor.

  Processor: PIC16F684
--------------------------------------------------------------------*/                
#include <system.h>

#pragma CLOCK_FREQ      4000000

//#pragma DATA _CONFIG, _FCMEN_OFF & _IESO_OFF & _BOD_OFF & _WDT_OFF & _INTRC_OSC_NOCLK 
// Config Bits: 11 0000 1110 0100
// bit 13-12: unimplemented, set to 1
// bit 11:    FCMEN  0 = fail safe clock disabled
// bit 10:    IESO   0 = internal external switchover disabled
// bits 9-8:  BODEN  00 = Brown out detect disabled
// bit 7:     !CPD   1 = data memory code protection disabled
// bit 6:     !CP    1 = program memory code protection disabled
// bit 5:     MCLRE  1 = RA3/!MCLR pin function is !MCLR
// bit 4:     ^PWRTE 0 = PWRT enabled
// bit 3:     WDTE   0 = watchdog timer disabled
// bits 2-0:  FOSC   100 = INTOSC oscillator, I/O on RA4/OSC2/CLKOUT, I/O on RA5/OSC1/CLKIN

#pragma DATA _CONFIG, 0x30e4

#define TIMER1_H	0xeb			// TMR1H initial value

#define READINGS	10		// number of readings to average

// Global Variables

static unsigned char ucDigit=0;			// current digit to display
static unsigned char Ones=0, Tens=0, Hund=0;
unsigned char neg_flag;

static unsigned char ucReadings=0;
static unsigned int uiTable[READINGS];

// Internal Functions
static unsigned int averageTemp( unsigned int uiValue );
static void breakNumber( unsigned int uiNumber );
static unsigned int convertTemp( unsigned int uiAdc0 );
static void initCPU( void );
static void timer1Overflow( void );

/*---------------------------------------------------------------------
  main()
---------------------------------------------------------------------*/
void main()
{
  unsigned int uiAdc0=0, uiTempC1, uiTempC2;

  initCPU();                	// initialize hardware

  delay_s( 1 );

  while ( 1 )
  {   
    // Sample AN0 (TC1046)
    set_bit( adcon0, GO );              // initiate A/D conversion
    while( test_bit( adcon0, GO ) );	// wait for A/D completion
    MAKESHORT( uiAdc0, adresl, adresh );
    
    uiTempC1 = convertTemp( uiAdc0 );
    uiTempC2 = averageTemp( uiTempC1 );
    breakNumber( uiTempC2 );				// display the result on the LED
    
    delay_s( 1 );
  }
}

/*---------------------------------------------------------------------
  convertTemp()

  Convert the A/D input value (0..1023 representing a voltage in the
  range 0..1 volts) to tenths of degrees C.
  
  The TC1046 output voltage range is typically 174 mv at -40C, 424 mV
  at 0C, 580 mV at 25C, 1205 mV at 125C. The slope is 6.25 mV per 
  degree C.
  
  So first we need to convert the A/D voltage to millivolts, then
  to tenths of degrees C.
  
  The formula for conversion is 
  
				  V - 424
			T = ------------
				   6.25
				   
  where T is temperature in degrees C, and V is Vout in millivolts.
  
  Note: we're not going to handle negatives for now!
  
    Input: A/D reading
  Returns: tens of degrees C
---------------------------------------------------------------------*/
static unsigned int convertTemp( unsigned int uiAdc0 )
{
  unsigned long ulVoutMv;
  unsigned int uiTempTensC;

  // Input is in the range 0..1023 which represents a voltage of 0..1
  // volts (Vref). Scale the adc value to a voltage value (millivolts).

  ulVoutMv = ( (unsigned long) uiAdc0 * 1000L ) / 1024L;

  // Derive temperature (tens of degrees C) from the formula above.
  // Multiply numerator and denominator by 100 to handle decimals.
  // Multiply numerator by 10 to get tens of degrees.
  
  uiTempTensC = (unsigned int) ( ( ( ulVoutMv - 424L ) * 1000L ) / 625L );

  return uiTempTensC;				// return tenths of degrees C
  //return (unsigned int) ulVoutMv;	// return Vout millivolts
  //return uiAdc0;					// return A/D reading
}

/*---------------------------------------------------------------------
  averageTemp(...)
  
  I think that the A/D converter is overstressed with this sensor.
  There is typically a jitter of +/- 1 in the A/D output, which
  translates into 0.6 degrees C. Occasionally the jitter is a bit
  greater, so the output reading jumps around almost one degree C.
  This is very annoying, so to correct the problem I try averaging.
  With READINGS=10, the jitter is greatly reduced to a change of
  .1 degree every few seconds. Perhaps there's a better way to 
  solve the problem?
  
  Averages READINGS numbers.
---------------------------------------------------------------------*/
static unsigned int averageTemp( unsigned int uiValue )
{
  unsigned char ii, jj, nn;
  unsigned int uiSum, uiAvg, uiRem;

  if ( !ucReadings )
  {
    // The 1st time, initialize the table
    for ( ii=0; ii<READINGS; ++ii )
      uiTable[ii] = uiValue;
  }
  
  // Shift all the table entries up one position
  nn = READINGS - 1;
  jj = 1;
  for ( ii=0; ii<nn; ++ii )
  {
    uiTable[ii] = uiTable[jj];
    ++jj;
  }  

  // Slot in the new reading at the end of the table
  uiTable[nn] = uiValue;
  if ( ucReadings < READINGS )
    ++ucReadings;
  
  // Take the average
  uiSum = 0;
  for ( ii=0; ii<READINGS; ++ii )
    uiSum += uiTable[ii];
    
  uiAvg = uiSum / READINGS;
  uiRem = uiSum % READINGS;
  if ( uiRem > ( READINGS / 2 ) )	// I hope this is rounding
    ++uiAvg;
  
  return uiAvg;
}

/*---------------------------------------------------------------------
  interrupt()
---------------------------------------------------------------------*/
void interrupt( void )
{
  if ( test_bit( pir1, TMR1IF ) )
  {
    timer1Overflow();
    tmr1h = TIMER1_H;
    clear_bit( pir1, TMR1IF );    // reset the TMR1 overflow interrupt flag
  }  
}

/*---------------------------------------------------------------------
  timer1Overflow()
  
  Timer1 is used to display digits on the LED. There have 3 digits, and
  each digit must be displayed in sequence, fast enough that the display
  appears fixed to the user.
  
  Timer1 counts down 16 bits, giving a maximum of 65536 counts. Counting
  is done by the processor, each instruction cycle. Since the clock is
  4 mHz, this gives an instruction cycle of 1 mHz or 1 microsecond. So
  it takes 65536 us or 65.536 ms to count up. There is also a pre-
  scaler available, from 1:1 to 8:1, so the resulting countdown times
  are: 65 ms, 131.072 ms, 262.144 ms, 524,288 ms.
  
  To get the LED display to not flicker, it must be updated fast. The
  prescaler is not required and set to unity. TMR1H, the high order
  byte of the counter, is loaded with a value determined experimentally
  to be just sufficient to prevent flicker, around 5 ms per digit.
---------------------------------------------------------------------*/
static void timer1Overflow( void )
{
  unsigned char cDigit;

  if ( ucDigit == 2 )
    cDigit = Ones;
  else
    if ( ucDigit == 1 )
      cDigit = ( Tens || Hund ) ? Tens : 15;
    else
      cDigit = ( Hund ) ? Hund : 15;    // 15 suppresses leading zero

  // All digits off

  set_bit( porta, 2 );          // RA2 on, digit 1 off
  set_bit( porta, 4 );          // RA4 on, digit 2 off
  set_bit( porta, 5 );          // RA5 on, digit 3 off
  set_bit( portc, 4 );          // RC4 on, decimal point off
  
  // Output the digit as BCD on RC0..RC3
  
  if ( test_bit( cDigit, 0 ) )
    set_bit( portc, 0 );
  else
    clear_bit( portc, 0 );
  
  if ( test_bit( cDigit, 1 ) )
    set_bit( portc, 1 );
  else
    clear_bit( portc, 1 );
  
  if ( test_bit( cDigit, 2 ) )
    set_bit( portc, 2 );
  else
    clear_bit( portc, 2 );
  
  if ( test_bit( cDigit, 3 ) )
    set_bit( portc, 3 );
  else
    clear_bit( portc, 3 );

  // Select the digit

  if ( ucDigit == 2 )
  {
    clear_bit( porta, 5 );		// RA5 off, digit 3 on
    ucDigit = 0;
  }  
  else
    if ( ucDigit == 1 )
    {
      clear_bit( porta, 4 );	// RA4 off, digit 2 on
      clear_bit( portc, 4 );    // RC4 off, decimal point on
      ucDigit = 2;
    }  
    else
    {
      clear_bit( porta, 2 );	// RA2 off, digit 1 on
      ucDigit = 1;
    }  
}

/*---------------------------------------------------------------------
  breakNumber(...)

  Breaks the passed number into digits Hund, Tens, Ones.
---------------------------------------------------------------------*/
static void breakNumber( unsigned int uiNumber )
{
  unsigned int uiTemp;

  if ( uiNumber > 999 )
  {
    Hund = Tens = Ones = 15;    // error, blank the display
  }
  else
  {
    Hund = (unsigned char) ( uiNumber / 100 );
    uiTemp = uiNumber - ( (unsigned) Hund * 100 );
    Tens = (unsigned char) ( uiTemp / 10 );
    Ones = (unsigned char) ( uiTemp - ( (unsigned) Tens * 10 ) );
  }
}
/*---------------------------------------------------------------------
  initCPU()
---------------------------------------------------------------------*/
static void initCPU()
{

  // Initialize the internal clock:
  osccon = 01100000b;		// internal clock @ 4MHz, use FOSC (config)

  // PORTC connections as follows:
  //  RC0 - (out) LED segment bit 0
  //  RC1 - (out) LED segment bit 1
  //  RC2 - (out) LED segment bit 2
  //  RC3 - (out) LED segment bit 3
  //  RC4 - (out) LED decimal point
  //  RC5 - (out) n/c

  portc = 0;
  trisc = 00000000b;            // 1=input, 0=output
  cmcon0 = 00000111b;			// comparitors off

  // PORTA connections as follows:
  //  RA0 - (in) A/D (AD0)
  //  RA1 - (in) A/D (AD1)
  //  RA2 - (out) LED digit 1 anode control
  //  RA3 - (in) !MCLR
  //  RA4 - (out) LED digit 2 anode control
  //  RA5 - (out) LED digit 3 anode control

  trisa = 00001011b;			// 1=input, 0=output
  porta = 00011010b;			// LED de-select bits on

  // Initialize the A/D
  ansel = 00000011b;			// AN0, AN1 are analog, others digital
  adcon0 = 11000001b;			// right justify; ref=Vref; channel=AN0 
  adcon1 = 00010000b;			// clock = Fosc/8

  // Timer1 initialization

  tmr1l = 0;
  tmr1h = TIMER1_H;
  t1con = 00000001b;            // 1:1 prescale; int. clock; timer enabled

  // Enable interrupts for ADC, Timer1, USART Rx

  set_bit( intcon, GIE );       // enable global interrupts
  set_bit( intcon, PEIE );      // enable peripheral interrupts

  set_bit( pie1, TMR1IE );      // Timer 1 interrupts
}
